#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

from ..data import check_input
from ..data import check_input_multi
from ..data import create_crosslink_from_csm
from ..data import create_parser_result
from .util import get_available_keys

from typing import Dict
from typing import List
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __score_better(
    score: float, reference: float, function: Literal["higher_better", "lower_better"]
) -> bool:
    r"""Checks if the score is better than the provided reference score.

    Checks if the score is better than the provided reference score using the given scoring scheme.

    Parameters
    ----------
    score : float
        The score that should be compared.
    reference : float
        The reference score to compare to.
    function : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    bool
        If the given score is better than the reference score.
    """
    if function == "higher_better":
        return score > reference
    return score < reference


def __get_csm_key(csm: Dict[str, Any]) -> str:
    r"""Get the unique key for a crosslink-spectrum-match.

    Parameters
    ----------
    csm : dict of str, any
        A pyXLMS crosslink-spectrum-match object.

    Returns
    -------
    str
        The unique key for the crosslink-spectrum-match.
    """
    return f"{csm['spectrum_file']}_{csm['scan_nr']}"


def __get_xl_key(xl: Dict[str, Any], by: Literal["peptide", "protein"]) -> str:
    r"""Get the unique key for a crosslink.

    Parameters
    ----------
    xl : dict of str, any
        A pyXLMS crosslink object.
    by : str, one of "peptide" or "protein"
        If peptide or protein crosslink position should be used for determining if a crosslink is unique.

    Returns
    -------
    str
        The unique key for the crosslink.

    Notes
    -----
    This function should not be called directly, it is called from ``__unique_xls()``.
    """
    if by == "peptide":
        return f"{xl['alpha_peptide']}_{xl['alpha_peptide_crosslink_position']}-{xl['beta_peptide']}_{xl['beta_peptide_crosslink_position']}"
    prot_pos_a = "-".join(
        sorted(
            [
                f"{xl['alpha_proteins'][i]}_{xl['alpha_proteins_crosslink_positions'][i]}"
                for i in range(len(xl["alpha_proteins"]))
            ]
        )
    )
    prot_pos_b = "-".join(
        sorted(
            [
                f"{xl['beta_proteins'][i]}_{xl['beta_proteins_crosslink_positions'][i]}"
                for i in range(len(xl["beta_proteins"]))
            ]
        )
    )
    return ":".join(sorted([prot_pos_a, prot_pos_b]))


def __unique_csms(
    csms: List[Dict[str, Any]],
    has_scores: bool,
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Filter for unique crosslink-spectrum-matches from a list on non-unique crosslink-spectrum-matches.

    Filters for unique crosslink-spectrum-matches from a list on non-unique crosslink-spectrum-matches. A crosslink-
    spectrum-match is considered unique if there is no other crosslink-spectrum-match from the same spectrum file and
    with the same scan number. If more than one crosslink-spectrum-match exists per spectrum file and scan number, the
    one with the better/best score is kept and the rest is filtered out. If crosslink-spectrum-matches without scores
    are provided, the first crosslink-spectrum-match in the list is kept instead.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of pyXLMS crosslink-spectrum-match objects.
    has_scores : bool
        If the crosslink-spectrum-match objects contain scores.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        List of unique crosslink-spectrum-matches.

    Notes
    -----
    This function should not be called directly, it is called from ``unique()``.
    """
    unique_csms = dict()
    for csm in csms:
        key = __get_csm_key(csm)
        if key not in unique_csms:
            unique_csms[key] = csm
        elif has_scores and __score_better(
            csm["score"], unique_csms[key]["score"], score
        ):
            unique_csms[key] = csm
        else:
            # do nothing
            pass
    return list(unique_csms.values())


def __unique_xls(
    xls: List[Dict[str, Any]],
    by: Literal["peptide", "protein"],
    has_scores: bool,
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Filter for unique crosslinks from a list on non-unique crosslinks.

    Filters for unique crosslinks from a list on non-unique crosslinks. A crosslink is considered unique if there is no
    other crosslink with the same peptide sequence and crosslink position if ``by = "peptide"``, otherwise it is considered
    unique if there are no other crosslinks with the same protein crosslink position (residue pair). If more than one
    crosslink exists per peptide sequence/residue pair, the one with the better/best score is kept and the rest is filtered
    out. If crosslinks without scores are provided, the first crosslink in the list is kept instead.

    Parameters
    ----------
    xls : list of dict of str, any
        A list of pyXLMS crosslink objects.
    by : str, one of "peptide" or "protein"
        If peptide or protein crosslink position should be used for determining if a crosslink is unique.
    has_scores : bool
        If the crosslink objects contain scores.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        List of unique crosslinks.

    Notes
    -----
    This function should not be called directly, it is called from ``unique()``.
    """
    unique_xls = dict()
    for xl in xls:
        key = __get_xl_key(xl, by)
        if key not in unique_xls:
            unique_xls[key] = xl
        elif has_scores and __score_better(
            xl["score"], unique_xls[key]["score"], score
        ):
            unique_xls[key] = xl
        else:
            # do nothing
            pass
    return list(unique_xls.values())


def unique(
    data: List[Dict[str, Any]] | Dict[str, Any],
    by: Literal["peptide", "protein"] = "peptide",
    score: Literal["higher_better", "lower_better"] = "higher_better",
) -> List[Dict[str, Any]] | Dict[str, Any]:
    r"""Filter for unique crosslinks or crosslink-spectrum-matches.

    Filters for unique crosslinks from a list on non-unique crosslinks. A crosslink is considered unique if there is no
    other crosslink with the same peptide sequence and crosslink position if ``by = "peptide"``, otherwise it is considered
    unique if there are no other crosslinks with the same protein crosslink position (residue pair). If more than one
    crosslink exists per peptide sequence/residue pair, the one with the better/best score is kept and the rest is filtered
    out. If crosslinks without scores are provided, the first crosslink in the list is kept instead.

    *or*

    Filters for unique crosslink-spectrum-matches from a list on non-unique crosslink-spectrum-matches. A crosslink-
    spectrum-match is considered unique if there is no other crosslink-spectrum-match from the same spectrum file and
    with the same scan number. If more than one crosslink-spectrum-match exists per spectrum file and scan number, the
    one with the better/best score is kept and the rest is filtered out. If crosslink-spectrum-matches without scores
    are provided, the first crosslink-spectrum-match in the list is kept instead.

    Parameters
    ----------
    data : dict of str, any, or list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to filter, or a parser_result.
    by : str, one of "peptide" or "protein", default = "peptide"
        If peptide or protein crosslink position should be used for determining if a crosslink is unique.
        Only affects filtering for unique crosslinks and not crosslink-spectrum-matches. If protein crosslink
        position is not available for all crosslinks a ``ValueError`` will be raised. Make sure that all
        crosslinks have the ``_proteins`` and ``_proteins_crosslink_positions`` fields set. If this is not
        already done by the parser, this can be achieved with ``transform.reannotate_positions()``.
    score : str, one of "higher_better" or "lower_better", default = "higher_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any, or dict of str, any
        If a list of crosslink-spectrum-matches or crosslinks was provided, a list of unique
        crosslink-spectrum-matches or crosslinks is returned. If a parser_result was provided,
        a parser_result with unique crosslink-spectrum-matches and/or unique crosslinks will
        be returned.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If parameter by is not one of 'peptide' or 'protein'.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    ValueError
        If parameter by is set to 'protein' but protein crosslink positions are not available.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import unique
    >>> pr = read(
    ...     ["data/_test/aggregate/csms.txt", "data/_test/aggregate/xls.txt"],
    ...     engine="custom",
    ...     crosslinker="DSS",
    ... )
    >>> len(pr["crosslink-spectrum-matches"])
    10
    >>> len(pr["crosslinks"])
    10
    >>> unique_peptide = unique(pr, by="peptide")
    >>> len(unique_peptide["crosslink-spectrum-matches"])
    5
    >>> len(unique_peptide["crosslinks"])
    3

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import unique
    >>> pr = read(
    ...     ["data/_test/aggregate/csms.txt", "data/_test/aggregate/xls.txt"],
    ...     engine="custom",
    ...     crosslinker="DSS",
    ... )
    >>> len(pr["crosslink-spectrum-matches"])
    10
    >>> len(pr["crosslinks"])
    10
    >>> unique_protein = unique(pr, by="protein")
    >>> len(unique_protein["crosslink-spectrum-matches"])
    5
    >>> len(unique_protein["crosslinks"])
    2
    """
    _ok = check_input_multi(data, "data", [dict, list])
    _ok = check_input(by, "by", str)
    _ok = check_input(score, "score", str)
    if by not in ["peptide", "protein"]:
        raise TypeError(
            "Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
            "peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
        )
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
            "-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
        )
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if len(data) == 0:
            return data
        if "data_type" not in data[0] or data[0]["data_type"] not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]:
            raise TypeError(
                "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match, "
                "or a parser_result!"
            )
        available_keys = get_available_keys(data)
        unique_items = list()
        if data[0]["data_type"] == "crosslink" and by == "protein":
            if (
                available_keys["alpha_proteins"]
                and available_keys["alpha_proteins_crosslink_positions"]
                and available_keys["beta_proteins"]
                and available_keys["beta_proteins_crosslink_positions"]
            ):
                unique_items += __unique_xls(data, by, available_keys["score"], score)
            else:
                raise ValueError(
                    "Grouping by protein crosslink position is only available if all crosslinks have defined protein crosslink positions!\n"
                    "This error might be fixable with 'transform.reannotate_positions()'!"
                )
        elif data[0]["data_type"] == "crosslink":
            unique_items += __unique_xls(data, by, available_keys["score"], score)
        else:
            unique_items += __unique_csms(data, available_keys["score"], score)
        return unique_items
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError(
            "Can't annotate positions for dict. Dict has to be a valid 'parser_result'!"
        )
    new_csms = (
        unique(data["crosslink-spectrum-matches"], by, score)
        if data["crosslink-spectrum-matches"] is not None
        else None
    )
    new_xls = (
        unique(data["crosslinks"], by, score)
        if data["crosslinks"] is not None
        else None
    )
    if new_csms is not None:
        if not isinstance(new_csms, list):
            raise RuntimeError(
                "Something went wrong while getting unique crosslink-spectrum-matches.\n"
                f"Expected data type: list. Got: {type(new_csms)}."
            )
    if new_xls is not None:
        if not isinstance(new_xls, list):
            raise RuntimeError(
                "Something went wrong while getting unique crosslinks.\n"
                f"Expected data type: list. Got: {type(new_xls)}."
            )
    return create_parser_result(
        search_engine=data["search_engine"], csms=new_csms, crosslinks=new_xls
    )


def aggregate(
    csms: List[Dict[str, Any]],
    by: Literal["peptide", "protein"] = "peptide",
    score: Literal["higher_better", "lower_better"] = "higher_better",
) -> List[Dict[str, Any]]:
    r"""Aggregate crosslink-spectrum-matches to crosslinks.

    Aggregates a list of crosslink-spectrum-matches to unique crosslinks. A crosslink is considered unique if there is no
    other crosslink with the same peptide sequence and crosslink position if ``by = "peptide"``, otherwise it is considered
    unique if there are no other crosslinks with the same protein crosslink position (residue pair). If more than one
    crosslink exists per peptide sequence/residue pair, the one with the better/best score is kept and the rest is filtered
    out. If crosslink-spectrum-matches without scores are provided, the crosslink of the first corresponding crosslink-spectrum
    -match in the list is kept instead.

    Parameters
    ----------
    csms : list of dict of str, any
        A list of crosslink-spectrum-matches.
    by : str, one of "peptide" or "protein", default = "peptide"
        If peptide or protein crosslink position should be used for determining if a crosslink is unique.
        If protein crosslink position is not available for all crosslink-spectrum-matches a ``ValueError``
        will be raised. Make sure that all crosslink-spectrum-matches have the ``_proteins`` and
        ``_proteins_crosslink_positions`` fields set. If this is not already done by the parser, this can
        be achieved with ``transform.reannotate_positions()``.
    score : str, one of "higher_better" or "lower_better", default = "higher_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        A list of aggregated, unique crosslinks.

    Warnings
    --------
    Aggregation will not conserve false discovery rate (FDR)! Aggregating crosslink-spectrum-matches that are
    validated for 1% FDR will not result in crosslinks validated for 1% FDR! Aggregated crosslinks should be
    validated with either external tools or with the built-in ``transform.validate()``!

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If parameter by is not one of 'peptide' or 'protein'.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    ValueError
        If parameter by is set to 'protein' but protein crosslink positions are not available.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import aggregate
    >>> pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    >>> len(pr["crosslink-spectrum-matches"])
    10
    >>> aggregate_peptide = aggregate(pr["crosslink-spectrum-matches"], by="peptide")
    >>> len(aggregate_peptide)
    3

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import aggregate
    >>> pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    >>> len(pr["crosslink-spectrum-matches"])
    10
    >>> aggregate_protein = aggregate(pr["crosslink-spectrum-matches"], by="protein")
    >>> len(aggregate_protein)
    2
    """
    _ok = check_input(csms, "csms", list, dict)
    _ok = check_input(by, "by", str)
    _ok = check_input(score, "score", str)
    if by not in ["peptide", "protein"]:
        raise TypeError(
            "Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
            "peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
        )
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
            "-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
        )
    if len(csms) == 0:
        return csms
    if "data_type" not in csms[0] or csms[0]["data_type"] != "crosslink-spectrum-match":
        raise TypeError(
            "Unsupported data type for input csms! Parameter csms has to be a list of crosslink-spectrum-matches!"
        )
    available_keys = get_available_keys(csms)
    if by == "protein":
        if (
            available_keys["alpha_proteins"]
            and available_keys["alpha_proteins_crosslink_positions"]
            and available_keys["beta_proteins"]
            and available_keys["beta_proteins_crosslink_positions"]
        ):
            # all fine
            pass
        else:
            raise ValueError(
                "Grouping by protein crosslink position is only available if all crosslink-spectrum-matches have defined protein "
                "crosslink positions!\nThis error might be fixable with 'transform.reannotate_positions()'!"
            )
    xls = [create_crosslink_from_csm(csm) for csm in csms]
    aggregated = unique(xls, by, score)
    if not isinstance(aggregated, list):
        raise RuntimeError(
            "Something went wrong while aggregating crosslinks.\n"
            f"Expected data type: list. Got: {type(aggregated)}."
        )
    return aggregated



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.exporter import to_xmas

    with pytest.raises(
        ValueError,
        match="Provided crosslinks contain no elements!",
    ):
        _df = to_xmas([], filename=None)


def test2():
    from pyXLMS.exporter import to_xmas
    from pyXLMS.data import create_csm_min
    from pyXLMS.data import create_crosslink_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    xl1 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    mixed = [xl1, csm1]
    with pytest.raises(
        TypeError, match="Not all elements in data have the same data type!"
    ):
        _df = to_xmas(mixed, filename=None)


def test3():
    from pyXLMS.exporter import to_xmas
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]
    df = to_xmas(crosslinks, filename="crosslinks_xmas.xlsx")
    assert df.shape[0] == 2
    assert df.shape[1] == 2


def test4():
    from pyXLMS.exporter import to_xmas
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]
    df = to_xmas(crosslinks, filename=None)
    assert df.shape[0] == 2
    assert df.shape[1] == 2


def test5():
    from pyXLMS.exporter import to_xmas
    from pyXLMS.parser import read

    pr = read(
        "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv",
        engine="xiSearch/xiFDR",
        crosslinker="DSS",
    )
    crosslinks = pr["crosslinks"]
    df = to_xmas(crosslinks, filename=None)
    assert df.shape[0] == 227
    assert df.shape[1] == 2


def test6():
    from pyXLMS.exporter import to_xmas
    from pyXLMS.parser import read

    pr = read(
        "data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv",
        engine="xiSearch/xiFDR",
        crosslinker="DSS",
    )
    csms = pr["crosslink-spectrum-matches"]
    with pytest.raises(
        TypeError,
        match="Unsupported data type for input crosslinks! Parameter crosslinks has to be a list of crosslinks!",
    ):
        _df = to_xmas(csms, filename=None)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd
from tqdm import tqdm
from os.path import splitext

from ..data import check_input
from ..data import create_crosslink
from ..data import create_csm
from ..data import create_parser_result
from .util import format_sequence
from .util import get_bool_from_value
from .util import __serialize_pandas_series

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import Tuple
from typing import List
from typing import Callable

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def pyxlms_modification_str_parser(modifications: str) -> Dict[int, Tuple[str, float]]:
    r"""Parse a pyXLMS modification string.

    Parses a pyXLMS modification string and returns the pyXLMS specific modification object,
    a dictionary that maps positions to their modififications.

    Parameters
    ----------
    modifications : str
        The pyXLMS modification string.

    Returns
    -------
    dict of int, tuple
        The pyXLMS specific modification object, a dictionary that maps positions (1-based)
        to their respective modifications given as tuples of modification name and modification
        delta mass.

    Raises
    ------
    RuntimeError
        If multiple modifications on the same residue are parsed.

    Examples
    --------
    >>> from pyXLMS.parser import pyxlms_modification_str_parser
    >>> modification_str = "(1:[DSS|138.06808])"
    >>> pyxlms_modification_str_parser(modification_str)
    {1: ('DSS', 138.06808)}

    >>> from pyXLMS.parser import pyxlms_modification_str_parser
    >>> modification_str = "(1:[DSS|138.06808]);(7:[Oxidation|15.994915])"
    >>> pyxlms_modification_str_parser(modification_str)
    {1: ('DSS', 138.06808), 7: ('Oxidation', 15.994915)}
    """
    parsed_modifications = dict()
    for mod in modifications.split(";"):
        pos = int(mod.split("(")[1].split(":")[0])
        desc = mod.split("[")[1].split("|")[0].strip()
        mass = float(mod.split("|")[1].split("]")[0])
        # if this is really in pyXLMS format we don't need to check
        # if pos already exists, because that is impossible
        # but if the parser is used for other formats that recreate the
        # same modification representation we should maybe check?
        if pos in parsed_modifications:
            raise RuntimeError(f"Modification at position {pos} already exists!")
        parsed_modifications[pos] = (desc, mass)
    return parsed_modifications


def __get_value(row: pd.Series, column: str) -> Any | None:
    r"""Get value from column if it exists and is not None.

    Parameters
    ----------
    row : pd.Series
        A row from a pandas DataFrame.
    column : str
        The column name to be accessed.

    Returns
    -------
    any, or None
        The column value if it exists and is not None.

    Notes
    -----
    This function should not be called directly, it is called from ``read_custom()``.
    """
    if column not in row:
        return None
    if (
        pd.isna(row[column])
        or row[column] is None
        or str(row[column]).lower().strip() in ["", "nan", "null", "none"]
    ):  # pyright: ignore [reportGeneralTypeIssues]
        return None
    return row[column]


def read_custom(
    files: str | List[str] | BinaryIO,
    column_mapping: Optional[Dict[str, str]] = None,
    parse_modifications: bool = True,
    modification_parser: Optional[Callable[[str], Dict[int, Tuple[str, float]]]] = None,
    decoy_prefix: str = "REV_",
    format: Literal["auto", "csv", "txt", "tsv", "parquet", "xlsx"] = "auto",
    sep: str = ",",
    decimal: str = ".",
) -> Dict[str, Any]:
    r"""Read a custom or pyXLMS result file.

    Reads a custom or pyXLMS crosslink-spectrum-matches result file or crosslink result file in ``.csv``, ``.parquet``, or ``.xlsx`` format,
    and returns a ``parser_result``.

    The minimum required columns for a crosslink-spectrum-matches result file are:

    - "Alpha Peptide": The unmodified amino acid sequence of the first peptide.
    - "Alpha Peptide Crosslink Position": The position of the crosslinker in the sequence of the first peptide (1-based).
    - "Beta Peptide": The unmodified amino acid sequence of the second peptide.
    - "Beta Peptide Crosslink Position": The position of the crosslinker in the sequence of the second peptide (1-based).
    - "Spectrum File": Name of the spectrum file the crosslink-spectrum-match was identified in.
    - "Scan Nr": The corresponding scan number of the crosslink-spectrum-match.

    The minimum required columns for crosslink result file are:

    - "Alpha Peptide": The unmodified amino acid sequence of the first peptide.
    - "Alpha Peptide Crosslink Position": The position of the crosslinker in the sequence of the first peptide (1-based).
    - "Beta Peptide": The unmodified amino acid sequence of the second peptide.
    - "Beta Peptide Crosslink Position": The position of the crosslinker in the sequence of the second peptide (1-based).

    A full specification of columns that can be parsed can be found in the
    `docs <https://github.com/hgb-bin-proteomics/pyXLMS/blob/master/docs/format.md>`_.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the result file(s) or a file-like object/stream.
    column_mapping : dict of str, str
        A dictionary that maps the result file columns to the required pyXLMS column names.
    parse_modifications : bool, default = True
        Whether or not post-translational-modifications should be parsed for crosslink-spectrum-matches.
        Requires correct specification of the 'modification_parser' parameter.
    modification_parser : callable, or None
        A function that parses modification strings and returns the pyXLMS specific modifications object.
        If None, the function ``pyxlms_modification_str_parser()`` is used. If no modification columns are
        given this parameter is ignored.
    decoy_prefix : str, default = "REV\_"
        The prefix that indicates that a protein is from the decoy database.
    format : "auto", "csv", "tsv", "txt", "parquet", or "xlsx", default = "auto"
        The format of the result file. ``"auto"`` is only available if the name/path to the result file is given.
    sep : str, default = ","
        Seperator used in the ``.csv`` or ``.tsv`` file. Parameter is ignored if the file is in ``.xlsx`` format.
    decimal : str, default = "."
        Character to recognize as decimal point. Parameter is ignored if the file is in ``.xlsx`` format.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    ValueError
        If the input format is not supported or cannot be inferred.
    TypeError
        If one of the values could not be parsed.
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslinks or crosslink-spectrum-matches.

    Examples
    --------
    >>> from pyXLMS.parser import read_custom
    >>> csms_from_pyxlms = read_custom("data/pyxlms/csm.txt")

    >>> from pyXLMS.parser import read_custom
    >>> crosslinks_from_pyxlms = read_custom("data/pyxlms/xl.txt")
    """
    ## check input
    _ok = (
        check_input(column_mapping, "column_mapping", dict, str)
        if column_mapping is not None
        else True
    )
    _ok = check_input(parse_modifications, "parse_modifications", bool)
    _ok = (
        check_input(modification_parser, "modification_parser", Callable)
        if modification_parser is not None
        else True
    )
    _ok = check_input(decoy_prefix, "decoy_prefix", str)
    _ok = check_input(format, "format", str)
    _ok = check_input(sep, "sep", str)
    _ok = check_input(decimal, "decimal", str)
    ## helper functions

    def get_is_decoy_value(
        row: pd.Series, decoy_prefix: str, alpha: bool
    ) -> bool | None:
        if alpha:
            if __get_value(row, "Alpha Decoy") is not None:
                return get_bool_from_value(__get_value(row, "Alpha Decoy"))
            if __get_value(row, "Alpha Proteins") is not None:
                return decoy_prefix in str(__get_value(row, "Alpha Proteins"))
            return None
        if __get_value(row, "Beta Decoy") is not None:
            return get_bool_from_value(__get_value(row, "Beta Decoy"))
        if __get_value(row, "Beta Proteins") is not None:
            return decoy_prefix in str(__get_value(row, "Beta Proteins"))
        return None

    def get_int(value: Any) -> int | None:
        if value is None:
            return None
        try:
            return int(value)
        except Exception as _e:
            pass
        raise TypeError(f"Could not parse int from value {value}!")
        return None

    def get_float(value: Any) -> float | None:
        if value is None:
            return None
        try:
            return float(value)
        except Exception as _e:
            pass
        raise TypeError(f"Could not parse float from value {value}!")
        return None

    ## set default parser
    if modification_parser is None:
        modification_parser = pyxlms_modification_str_parser

    ## data structures
    crosslinks = list()
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    for input in inputs:
        ## reading data
        data = None
        if format == "auto" and not isinstance(input, str):
            raise ValueError(
                "Can't detect format for file-like objects. Please specify format manually!"
            )
        # and isinstance specified for type checking
        if format == "auto" and isinstance(input, str):
            file_extension = splitext(input)[1].lower()
            if (
                file_extension == ".txt"
                or file_extension == ".tsv"
                or file_extension == ".csv"
            ):
                data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
            elif file_extension == ".parquet":
                data = pd.read_parquet(input)
            elif file_extension == ".xlsx":
                data = pd.read_excel(input, engine="openpyxl")
            else:
                raise ValueError(
                    f"Detected file extension {file_extension} is not supported! Input file has to be a valid file with extension '.csv', '.tsv', '.parquet' or '.xlsx'!"
                )
        elif format in ["csv", "tsv", "txt", "parquet", "xlsx"]:
            if format == "xlsx":
                data = pd.read_excel(input, engine="openpyxl")
            elif format == "parquet":
                data = pd.read_parquet(input)
            else:
                data = pd.read_csv(input, sep=sep, decimal=decimal, low_memory=False)
        else:
            raise ValueError(
                f"Provided input format {format} is not supported! Input format has to be of type 'csv', 'tsv', 'parquet' or 'xlsx'!"
            )
        if data is None:
            raise RuntimeError(
                "Something went wrong while reading the file! Please file a bug report!"
            )
        # this should be impossible, but check here for pyright
        if not isinstance(data, pd.DataFrame):
            raise RuntimeError(
                "Something went wrong while reading the file! Please file a bug report!"
            )
        ## detect input file type
        if column_mapping is not None:
            data.rename(columns=column_mapping, inplace=True)
        col_names = data.columns.values.tolist()
        is_crosslink_dataframe = "Scan Nr" not in col_names
        ## process data
        if is_crosslink_dataframe:
            for i, row in tqdm(
                data.iterrows(),
                total=data.shape[0],
                desc="Reading crosslinks...",
            ):
                # create crosslink
                crosslink = create_crosslink(
                    peptide_a=format_sequence(str(row["Alpha Peptide"])),
                    xl_position_peptide_a=int(row["Alpha Peptide Crosslink Position"]),
                    proteins_a=[
                        protein.strip()
                        if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein.strip()[len(decoy_prefix) :]
                        for protein in str(__get_value(row, "Alpha Proteins")).split(
                            ";"
                        )
                    ]
                    if __get_value(row, "Alpha Proteins") is not None
                    else None,
                    xl_position_proteins_a=[
                        int(position)
                        for position in str(
                            __get_value(row, "Alpha Proteins Crosslink Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Alpha Proteins Crosslink Positions")
                    is not None
                    else None,
                    decoy_a=get_is_decoy_value(row, decoy_prefix, True),
                    peptide_b=format_sequence(str(row["Beta Peptide"])),
                    xl_position_peptide_b=int(row["Beta Peptide Crosslink Position"]),
                    proteins_b=[
                        protein.strip()
                        if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein.strip()[len(decoy_prefix) :]
                        for protein in str(__get_value(row, "Beta Proteins")).split(";")
                    ]
                    if __get_value(row, "Beta Proteins") is not None
                    else None,
                    xl_position_proteins_b=[
                        int(position)
                        for position in str(
                            __get_value(row, "Beta Proteins Crosslink Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Beta Proteins Crosslink Positions") is not None
                    else None,
                    decoy_b=get_is_decoy_value(row, decoy_prefix, False),
                    score=get_float(__get_value(row, "Crosslink Score")),
                    additional_information={"source": __serialize_pandas_series(row)},
                )
                crosslinks.append(crosslink)
        else:
            for i, row in tqdm(
                data.iterrows(),
                total=data.shape[0],
                desc="Reading CSMs...",
            ):
                # create csm
                csm = create_csm(
                    peptide_a=format_sequence(str(row["Alpha Peptide"])),
                    modifications_a=modification_parser(
                        str(__get_value(row, "Alpha Peptide Modifications"))
                    )
                    if parse_modifications
                    and __get_value(row, "Alpha Peptide Modifications") is not None
                    else None,
                    xl_position_peptide_a=int(row["Alpha Peptide Crosslink Position"]),
                    proteins_a=[
                        protein.strip()
                        if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein.strip()[len(decoy_prefix) :]
                        for protein in str(__get_value(row, "Alpha Proteins")).split(
                            ";"
                        )
                    ]
                    if __get_value(row, "Alpha Proteins") is not None
                    else None,
                    xl_position_proteins_a=[
                        int(position)
                        for position in str(
                            __get_value(row, "Alpha Proteins Crosslink Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Alpha Proteins Crosslink Positions")
                    is not None
                    else None,
                    pep_position_proteins_a=[
                        int(position)
                        for position in str(
                            __get_value(row, "Alpha Proteins Peptide Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Alpha Proteins Peptide Positions") is not None
                    else None,
                    score_a=get_float(__get_value(row, "Alpha Score")),
                    decoy_a=get_is_decoy_value(row, decoy_prefix, True),
                    peptide_b=format_sequence(str(row["Beta Peptide"])),
                    modifications_b=modification_parser(
                        str(__get_value(row, "Beta Peptide Modifications"))
                    )
                    if parse_modifications
                    and __get_value(row, "Beta Peptide Modifications") is not None
                    else None,
                    xl_position_peptide_b=int(row["Beta Peptide Crosslink Position"]),
                    proteins_b=[
                        protein.strip()
                        if protein.strip()[: len(decoy_prefix)] != decoy_prefix
                        else protein.strip()[len(decoy_prefix) :]
                        for protein in str(__get_value(row, "Beta Proteins")).split(";")
                    ]
                    if __get_value(row, "Beta Proteins") is not None
                    else None,
                    xl_position_proteins_b=[
                        int(position)
                        for position in str(
                            __get_value(row, "Beta Proteins Crosslink Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Beta Proteins Crosslink Positions") is not None
                    else None,
                    pep_position_proteins_b=[
                        int(position)
                        for position in str(
                            __get_value(row, "Beta Proteins Peptide Positions")
                        ).split(";")
                    ]
                    if __get_value(row, "Beta Proteins Peptide Positions") is not None
                    else None,
                    score_b=get_float(__get_value(row, "Beta Score")),
                    decoy_b=get_is_decoy_value(row, decoy_prefix, False),
                    score=get_float(__get_value(row, "CSM Score")),
                    spectrum_file=str(row["Spectrum File"]).strip(),
                    scan_nr=int(row["Scan Nr"]),
                    charge=get_int(__get_value(row, "Precursor Charge")),
                    rt=get_float(__get_value(row, "Retention Time")),
                    im_cv=get_float(__get_value(row, "Ion Mobility")),
                    additional_information={"source": __serialize_pandas_series(row)},
                )
                csms.append(csm)
    ## check results
    if len(crosslinks) + len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches or crosslinks were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="Custom",
        csms=csms if len(csms) > 0 else None,
        crosslinks=crosslinks if len(crosslinks) > 0 else None,
    )



#!/usr/bin/env python3

# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

__all__ = [
    "modifications_to_str",
    "assert_data_type_same",
    "get_available_keys",
    "filter_target_decoy",
    "filter_proteins",
    "filter_protein_distribution",
    "filter_crosslink_type",
    "filter_peptide_pair_distribution",
    "summary",
    "unique",
    "aggregate",
    "validate",
    "to_proforma",
    "to_dataframe",
    "from_dataframe",
    "targets_only",
    "fasta_title_to_accession",
    "reannotate_positions",
    "intersection",
    "annotate_fdr",
]

from .util import modifications_to_str
from .util import assert_data_type_same
from .util import get_available_keys
from .filter import filter_target_decoy
from .filter import filter_proteins
from .filter import filter_protein_distribution
from .filter import filter_crosslink_type
from .filter import filter_peptide_pair_distribution
from .summary import summary
from .aggregate import unique
from .aggregate import aggregate
from .validate import validate
from .to_proforma import to_proforma
from .to_dataframe import to_dataframe
from .from_dataframe import from_dataframe
from .targets_only import targets_only
from .reannotate_positions import fasta_title_to_accession
from .reannotate_positions import reannotate_positions
from .intersection import intersection
from .annotate_fdr import annotate_fdr



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

PLINK2 = "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv"
PLINK3 = "data/plink3/Cas10_plus10_2025.04.07.filtered_cross-linked_spectra.csv"


def test1():
    from pyXLMS.parser import read_plink
    from pyXLMS.transform import modifications_to_str as mts

    pr = read_plink(PLINK2, parse_modifications=False, modifications={})
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 961

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "FDNLTKAER"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 6
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [906]
    assert csm["alpha_proteins_peptide_positions"] == [901]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "YDENDKLIR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 6
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [952]
    assert csm["beta_proteins_peptide_positions"] == [947]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("5.553153e-009"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 13098
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[118]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [575]
    assert csm["alpha_proteins_peptide_positions"] == [575]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [575]
    assert csm["beta_proteins_peptide_positions"] == [575]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("8.010787e-004"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 19140
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[951]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "GKSDNVPSEEVVK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 2
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [870]
    assert csm["alpha_proteins_peptide_positions"] == [869]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "VKYVTEGMR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [532]
    assert csm["beta_proteins_peptide_positions"] == [531]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("4.356499e-001"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 11060
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "LKYENEMALR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 2
    assert csm["alpha_proteins"] == ["sp|K1C15_SHEEP|"]
    assert csm["alpha_proteins_crosslink_positions"] == [192]
    assert csm["alpha_proteins_peptide_positions"] == [191]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "TKYNALK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["sp|CTRB_BOVIN|"]
    assert csm["beta_proteins_crosslink_positions"] == [145]
    assert csm["beta_proteins_peptide_positions"] == [144]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(float("4.868460e-001"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 11341
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test2():
    from pyXLMS.parser import read_plink
    from pyXLMS.transform import modifications_to_str as mts

    pr = read_plink(PLINK3, parse_modifications=False, modifications={})
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "pLink"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 3

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "LFKGHPETLEK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] == ["sp|K1C15_SHEEP|", "sp|MYG_HUMAN|"]
    assert csm["alpha_proteins_crosslink_positions"] == [192, 34]
    assert csm["alpha_proteins_peptide_positions"] == [190, 32]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "LKYENEMALR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["sp|K1C15_SHEEP|", "sp|MYG_HUMAN|"]
    assert csm["beta_proteins_crosslink_positions"] == [192, 34]
    assert csm["beta_proteins_peptide_positions"] == [191, 33]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(float("0.332623"))
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1"
    assert csm["scan_nr"] == 14421
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2024 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

XISEARCH = "data/xi/r1_Xi1.7.6.7.csv"
XIFDR_CSMS = "data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv"
XIFDR_LINKS = "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv"


def test1():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    pr = p.read_xi(XISEARCH, parse_modifications=False, modifications={}, verbose=0)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "xiSearch/xiFDR"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 4648

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "SDKNR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 3
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [866]
    assert csm["alpha_proteins_peptide_positions"] == [864]
    assert csm["alpha_score"] == pytest.approx(0.596154)
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "SDKNR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 3
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [866]
    assert csm["beta_proteins_peptide_positions"] == [864]
    assert csm["beta_score"] == pytest.approx(0.596154)
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(8.758549)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 2561
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[29]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "CCPSR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["KRA3_SHEEP"]
    assert csm["alpha_proteins_crosslink_positions"] == [1]
    assert csm["alpha_proteins_peptide_positions"] == [1]
    assert csm["alpha_score"] == pytest.approx(0.240385)
    assert csm["alpha_decoy"]
    assert csm["beta_peptide"] == "SKKLK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 3
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [213]
    assert csm["beta_proteins_peptide_positions"] == [211]
    assert csm["beta_score"] == pytest.approx(0.355769)
    assert csm["beta_decoy"]
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(1.468129)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 6482
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[184]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "EKVYFQFK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 2
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [394.0]
    assert csm["alpha_proteins_peptide_positions"] == [393]
    assert csm["alpha_score"] == pytest.approx(0.214286)
    assert csm["alpha_decoy"]
    assert csm["beta_peptide"] == "MLRNER"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["RETBP_HUMAN"]
    assert csm["beta_proteins_crosslink_positions"] == [1]
    assert csm["beta_proteins_peptide_positions"] == [1]
    assert csm["beta_score"] == pytest.approx(0.138462)
    assert csm["beta_decoy"]
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(-0.014849)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 8316
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "EYNSAMKR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 7
    assert csm["alpha_proteins"] == ["MYG_HUMAN"]
    assert csm["alpha_proteins_crosslink_positions"] == [13]
    assert csm["alpha_proteins_peptide_positions"] == [7]
    assert csm["alpha_score"] == pytest.approx(0.0)
    assert csm["alpha_decoy"]
    assert csm["beta_peptide"] == "LKGSPEDNEQKQLFVEQHK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 2
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [1250]
    assert csm["beta_proteins_peptide_positions"] == [1249]
    assert csm["beta_score"] == pytest.approx(0.126068)
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "inter"
    assert csm["score"] == pytest.approx(0.48559)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 27092
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test2():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    pr = p.read_xi(XIFDR_CSMS, parse_modifications=False, modifications={}, verbose=0)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "xiSearch/xiFDR"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 413

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [575]
    assert csm["alpha_proteins_peptide_positions"] == [575]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [575]
    assert csm["beta_proteins_peptide_positions"] == [575]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(27.268)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 19140
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[332]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "NPIDFLEAKGYK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 9
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [1189]
    assert csm["alpha_proteins_peptide_positions"] == [1181]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QWYKNK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 4
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [488]
    assert csm["beta_proteins_peptide_positions"] == [485]
    assert csm["beta_score"] is None
    assert csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(10.071)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 21757
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "QITKHVAQILDSR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 4
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [933]
    assert csm["alpha_proteins_peptide_positions"] == [930]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QQLPEKYK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 6
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [350]
    assert csm["beta_proteins_peptide_positions"] == [345]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(6.808)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 12190
    assert csm["charge"] == 5
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test3():
    from pyXLMS import parser as p

    pr = p.read_xi(XIFDR_LINKS, parse_modifications=False, modifications={}, verbose=0)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "xiSearch/xiFDR"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 227

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "VVDELVKVMGR"
    assert xl["alpha_peptide_crosslink_position"] == 7
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [753]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "VVDELVKVMGR"
    assert xl["beta_peptide_crosslink_position"] == 7
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [753]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(40.679)

    xl = xls[30]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "FDNLTKAER"
    assert xl["alpha_peptide_crosslink_position"] == 6
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [952]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "YDENDKLIR"
    assert xl["beta_peptide_crosslink_position"] == 6
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [906]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(25.616)

    xl = xls[219]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "NPIDFLEAKGYK"
    assert xl["alpha_peptide_crosslink_position"] == 9
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [488]
    assert xl["alpha_decoy"]
    assert xl["beta_peptide"] == "QWYKNK"
    assert xl["beta_peptide_crosslink_position"] == 4
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [1189]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(10.071)

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "LSKSR"
    assert xl["alpha_peptide_crosslink_position"] == 3
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [222]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LSKSR"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [222]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(9.619)


def test4():
    from pyXLMS import parser as p
    from pyXLMS.transform import modifications_to_str as mts

    pr = p.read_xi(
        [XIFDR_LINKS, XIFDR_CSMS],
        parse_modifications=False,
        modifications={},
        verbose=0,
    )
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "full"
    assert pr["search_engine"] == "xiSearch/xiFDR"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is not None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 413

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [575]
    assert csm["alpha_proteins_peptide_positions"] == [575]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [575]
    assert csm["beta_proteins_peptide_positions"] == [575]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(27.268)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 19140
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[332]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "NPIDFLEAKGYK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 9
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [1189]
    assert csm["alpha_proteins_peptide_positions"] == [1181]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QWYKNK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 4
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [488]
    assert csm["beta_proteins_peptide_positions"] == [485]
    assert csm["beta_score"] is None
    assert csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(10.071)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 21757
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "QITKHVAQILDSR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 4
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [933]
    assert csm["alpha_proteins_peptide_positions"] == [930]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QQLPEKYK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 6
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [350]
    assert csm["beta_proteins_peptide_positions"] == [345]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(6.808)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 12190
    assert csm["charge"] == 5
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    xls = pr["crosslinks"]
    assert len(xls) == 227

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "VVDELVKVMGR"
    assert xl["alpha_peptide_crosslink_position"] == 7
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [753]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "VVDELVKVMGR"
    assert xl["beta_peptide_crosslink_position"] == 7
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [753]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(40.679)

    xl = xls[30]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "FDNLTKAER"
    assert xl["alpha_peptide_crosslink_position"] == 6
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [952]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "YDENDKLIR"
    assert xl["beta_peptide_crosslink_position"] == 6
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [906]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(25.616)

    xl = xls[219]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "NPIDFLEAKGYK"
    assert xl["alpha_peptide_crosslink_position"] == 9
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [488]
    assert xl["alpha_decoy"]
    assert xl["beta_peptide"] == "QWYKNK"
    assert xl["beta_peptide_crosslink_position"] == 4
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [1189]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(10.071)

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "LSKSR"
    assert xl["alpha_peptide_crosslink_position"] == 3
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [222]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LSKSR"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [222]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(9.619)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings

from ..data import check_input
from .aggregate import __score_better
from .aggregate import __get_xl_key
from .aggregate import unique
from .util import get_available_keys

from typing import Dict
from typing import List
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __get_csm_key(csm: Dict[str, Any]) -> str:
    r"""Get the unique intersection key for a crosslink-spectrum-match.

    Parameters
    ----------
    csm : dict of str, any
        A pyXLMS crosslink-spectrum-match object.

    Returns
    -------
    str
        The unique intersection key for the crosslink-spectrum-match.
    """
    k1 = csm["alpha_peptide"]
    k2 = csm["alpha_peptide_crosslink_position"]
    k3 = csm["beta_peptide"]
    k4 = csm["beta_peptide_crosslink_position"]
    k5 = csm["spectrum_file"]
    k6 = csm["scan_nr"]
    return f"{k1}{k2}-{k3}{k4}_{k5}_{k6}"


def intersection(
    data_a: List[Dict[str, Any]],
    data_b: List[Dict[str, Any]],
    use: Literal["better_score", "data_a", "data_b"] = "better_score",
    by: Literal["peptide", "protein"] = "peptide",
    score: Literal["higher_better", "lower_better"] = "higher_better",
    verbose: Literal[0, 1, 2] = 1,
) -> List[Dict[str, Any]]:
    r"""Get the intersection of two lists of crosslinks.

    Returns the intersection of two lists of crosslinks (or crosslink-spectrum-matches). Crosslink intersection is calculated
    by creating unique keys based on peptide sequence and peptide crosslink position or based on protein crosslink position.
    For crosslink-spectrum-match intersection the intersection is calculated by creating unique keys based on peptide sequence
    and peptide crosslink position as well as the corresponding spectrum file and scan number. For any unique key in the intersection,
    either the best (by score) crosslink or crosslink-spectrum-match is returned, or the one from the first or second list, based
    on user preference.

    Parameters
    ----------
    data_a : list of dict of str, any
        List of crosslinks (or crosslink-spectrum-matches).
    data_b : list of dict of str, any
        List of crosslinks (or crosslink-spectrum-matches) to intersect with. Note that the data types for ``data_a`` and
        ``data_b`` have to be the same.
    use : str, one of "better_score", "data_a", or "data_b", default = "better_score"
        Which element to use for the returned intersection. Option "better_score" will return the element (crosslink or
        crosslink-spectrum-match) with the higher score, option "data_a" will return the element from ``data_a``, and
        option "data_b" will return the element from ``data_b``. Please note that attribute ``score`` needs to be
        available for all elements if "better_score" is selected.
    by : str, one of "peptide" or "protein", default = "peptide"
        If peptide or protein crosslink position should be used for determining if a crosslink is unique.
        Only affects filtering for unique crosslinks and not crosslink-spectrum-matches. If protein crosslink
        position is not available for all crosslinks a ``ValueError`` will be raised. Make sure that all
        crosslinks have the ``_proteins`` and ``_proteins_crosslink_positions`` fields set. If this is not
        already done by the parser, this can be achieved with ``transform.reannotate_positions()``.
    score: str, one of "higher_better" or "lower_better", default = "higher_better"
        If a higher score is considered better, or a lower score is considered better.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    list of dict of str, any
        The list of crosslinks or crosslink-spectrum-matches in the intersection.

    Raises
    ------
    TypeError
        If a wrong data type is provided for 'data_a' or 'data_b'.
    TypeError
        If 'data_a' and 'data_b' are not of the same data type.
    TypeError
        If parameter use is not one of 'better_score', 'data_a' or 'data_b'.
    TypeError
        If parameter by is not one of 'peptide' or 'protein'.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    TypeError
        If parameter verbose was not set correctly.
    ValueError
        If parameter use is set to 'better_score' but scores are not available.
    ValueError
        If parameter by is set to 'protein' but protein crosslink positions are not available.
    RuntimeError
        If crosslink-spectrum-matches are provided as data but verbose level is set to 2.

    Warns
    -----
    RuntimeWarning
        If crosslink-spectrum-matches are provided as data but verbose level is set to 1.

    Notes
    -----
    While technically the intersection of crosslink-spectrum-matches is supported, please be aware that this mostly only makes
    sense when searching the same mass spectrometry file several times, as the crosslink-spectrum-match intersection takes into
    account the spectrum files and scan numbers, e.g. two crosslinks with the same peptide sequences but different scan numbers will
    not be considered the same and therefore will not be in the intersection. If you want to get the intersection of crosslink-spectrum-matches
    using their crosslinked peptide sequences or protein positions, you can aggregate them first with ``transform.aggregate()`` or create
    crosslinks with ``data.create_crosslink_from_csm()``. Please also note that when intersecting all duplicates are removed and
    aggregated the same way as in ``transform.unique()``. To intersect more than two lists please repeatedly call ``intersection()``
    on itself.

    Examples
    --------
    >>> from pyXLMS.pipelines import pipeline
    >>> from pyXLMS.transform import aggregate
    >>> from pyXLMS.transform import intersection
    >>> msannika = pipeline(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> msannika = aggregate(msannika["crosslink-spectrum-matches"])
    >>> len(msannika)
    235
    >>> maxquant = pipeline(
    ...     "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    ... )
    >>> maxquant = aggregate(maxquant["crosslink-spectrum-matches"])
    >>> len(maxquant)
    226
    >>> crosslinks_intersection = intersection(msannika, maxquant)
    >>> len(crosslinks_intersection)
    206

    >>> from pyXLMS.pipelines import pipeline
    >>> from pyXLMS.transform import aggregate
    >>> from pyXLMS.transform import intersection
    >>> msannika = pipeline(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> msannika = aggregate(msannika["crosslink-spectrum-matches"])
    >>> len(msannika)
    235
    >>> maxquant = pipeline(
    ...     "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    ... )
    >>> maxquant = aggregate(maxquant["crosslink-spectrum-matches"])
    >>> len(maxquant)
    226
    >>> crosslinks_intersection = intersection(msannika, maxquant)
    >>> len(crosslinks_intersection)
    206
    >>> plink = pipeline(
    ...     "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
    ...     engine="pLink",
    ...     crosslinker="DSS",
    ... )
    >>> plink = aggregate(plink["crosslink-spectrum-matches"])
    >>> len(plink)
    252
    >>> crosslinks_intersection = intersection(crosslinks_intersection, plink)
    >>> len(crosslinks_intersection)
    203
    """
    _ok = check_input(data_a, "data_a", list, dict)
    _ok = check_input(data_b, "data_b", list, dict)
    _ok = check_input(use, "use", str)
    _ok = check_input(by, "by", str)
    _ok = check_input(score, "score", str)
    if use not in ["better_score", "data_a", "data_b"]:
        raise TypeError(
            "Parameter 'use' has to be one 'better_score', 'data_a', or 'data_b'! Option 'better_score' will return the intersection "
            "with the better score (if scores are available). Option 'data_a' will return the intersection using elements of 'data_a'. "
            "Option 'data_b' will return the intersection using elements of 'data_b'."
        )
    if by not in ["peptide", "protein"]:
        raise TypeError(
            "Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
            "peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
        )
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
            "-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
        )
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")
    if len(data_a) == 0 or len(data_b) == 0:
        return []
    if (
        "data_type" not in data_a[0]
        or data_a[0]["data_type"]
        not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]
        or "data_type" not in data_b[0]
        or data_b[0]["data_type"]
        not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]
    ):
        raise TypeError(
            "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match!"
        )
    if data_a[0]["data_type"] != data_b[0]["data_type"]:
        raise TypeError(
            "Parameters 'data_a' and 'data_b' have to be of the same data type!"
        )
    available_keys_a = get_available_keys(data_a)
    available_keys_b = get_available_keys(data_b)
    if use == "better_score" and (
        not available_keys_a["score"] or not available_keys_b["score"]
    ):
        raise ValueError(
            "Can't intersect based on score because not all data have associated scores!"
        )
    unique_a = unique(data_a, by=by, score=score)
    unique_b = unique(data_b, by=by, score=score)
    if not isinstance(unique_a, list) or not isinstance(unique_b, list):
        raise RuntimeError(
            "Something went wrong while getting unique data.\n"
            f"Expected data type: list. Got: {type(unique_a)} and {type(unique_b)}."
        )
    if unique_a[0]["data_type"] == "crosslink":
        crosslinks_a = dict()
        for xl in unique_a:
            crosslinks_a[__get_xl_key(xl, by=by)] = xl
        crosslinks_b = dict()
        for xl in unique_b:
            crosslinks_b[__get_xl_key(xl, by=by)] = xl
        keys_intersection = set(crosslinks_a.keys()).intersection(
            set(crosslinks_b.keys())
        )
        crosslinks_intersection = list()
        for key in keys_intersection:
            if use == "data_a":
                crosslinks_intersection.append(crosslinks_a[key])
            elif use == "data_b":
                crosslinks_intersection.append(crosslinks_b[key])
            else:
                if __score_better(
                    score=crosslinks_a[key]["score"],
                    reference=crosslinks_b[key]["score"],
                    function=score,
                ):
                    crosslinks_intersection.append(crosslinks_a[key])
                else:
                    crosslinks_intersection.append(crosslinks_b[key])
        return crosslinks_intersection
    if verbose == 1:
        warnings.warn(
            RuntimeWarning(
                "Creating intersection of crosslink-spectrum-matches. Be sure that this makes sense for your data!"
            )
        )
    elif verbose == 2:
        raise RuntimeError(
            "Can't create intersection of crosslink-spectrum-matches for verbose level 2!"
        )
    csms_a = dict()
    for csm in unique_a:
        csms_a[__get_csm_key(csm)] = csm
    csms_b = dict()
    for csm in unique_b:
        csms_b[__get_csm_key(csm)] = csm
    keys_intersection = set(csms_a.keys()).intersection(set(csms_b.keys()))
    csms_intersection = list()
    for key in keys_intersection:
        if use == "data_a":
            csms_intersection.append(csms_a[key])
        elif use == "data_b":
            csms_intersection.append(csms_b[key])
        else:
            if __score_better(
                score=csms_a[key]["score"],
                reference=csms_b[key]["score"],
                function=score,
            ):
                csms_intersection.append(csms_a[key])
            else:
                csms_intersection.append(csms_b[key])
    return csms_intersection



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.constants import AMINO_ACIDS

    assert "A" in AMINO_ACIDS
    assert "B" in AMINO_ACIDS


def test2():
    from pyXLMS.constants import AMINO_ACIDS_3TO1

    assert AMINO_ACIDS_3TO1["GLY"] == "G"


def test3():
    from pyXLMS.constants import AMINO_ACIDS_1TO3

    assert AMINO_ACIDS_1TO3["G"] == "GLY"


def test4():
    from pyXLMS.constants import CROSSLINKERS

    assert CROSSLINKERS["BS3"] == pytest.approx(138.06808)


def test5():
    from pyXLMS.constants import MODIFICATIONS

    assert MODIFICATIONS["Carbamidomethyl"] == pytest.approx(57.021464)
    assert MODIFICATIONS["BS3"] == pytest.approx(138.06808)


def test6():
    from pyXLMS.constants import XI_MODIFICATION_MAPPING

    assert XI_MODIFICATION_MAPPING["cm"][0] == "Carbamidomethyl"
    assert XI_MODIFICATION_MAPPING["cm"][1] == pytest.approx(57.021464)
    assert XI_MODIFICATION_MAPPING["ox"][0] == "Oxidation"
    assert XI_MODIFICATION_MAPPING["ox"][1] == pytest.approx(15.994915)


def test7():
    from pyXLMS.constants import SCOUT_MODIFICATION_MAPPING

    assert SCOUT_MODIFICATION_MAPPING["+57.021460"][0] == "Carbamidomethyl"
    assert SCOUT_MODIFICATION_MAPPING["+57.021460"][1] == pytest.approx(57.021464)
    assert SCOUT_MODIFICATION_MAPPING["Carbamidomethyl"][0] == "Carbamidomethyl"
    assert SCOUT_MODIFICATION_MAPPING["Carbamidomethyl"][1] == pytest.approx(57.021464)
    assert SCOUT_MODIFICATION_MAPPING["Oxidation of Methionine"][0] == "Oxidation"
    assert SCOUT_MODIFICATION_MAPPING["Oxidation of Methionine"][1] == pytest.approx(
        15.994915
    )


def test8():
    from pyXLMS.constants import MEROX_MODIFICATION_MAPPING

    assert MEROX_MODIFICATION_MAPPING["B"] == {
        "Amino Acid": "C",
        "Modification": ("Carbamidomethyl", 57.021464),
    }
    assert MEROX_MODIFICATION_MAPPING["m"] == {
        "Amino Acid": "M",
        "Modification": ("Oxidation", 15.994915),
    }
    assert MEROX_MODIFICATION_MAPPING == {
        "B": {"Amino Acid": "C", "Modification": ("Carbamidomethyl", 57.021464)},
        "m": {"Amino Acid": "M", "Modification": ("Oxidation", 15.994915)},
    }


def test9():
    from pyXLMS.constants import AMINO_ACIDS_REPLACEMENTS

    assert "X" in AMINO_ACIDS_REPLACEMENTS



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

from .parser import read
from .transform.summary import summary as transform_summary
from .transform.aggregate import unique as transform_unique
from .transform.validate import validate as transform_validate
from .transform.targets_only import targets_only as transform_targets_only

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import List

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def pipeline(
    files: str | List[str] | BinaryIO,
    engine: Literal[
        "Custom",
        "MaxQuant",
        "MaxLynx",
        "MeroX",
        "MS Annika",
        "mzIdentML",
        "pLink",
        "Scout",
        "xiSearch/xiFDR",
        "XlinkX",
    ],
    crosslinker: str,
    unique: Optional[bool | Dict[str, Any]] = True,
    validate: Optional[bool | Dict[str, Any]] = True,
    targets_only: Optional[bool] = True,
    **kwargs,
) -> Dict[str, Any]:
    r"""Runs a standard down-stream analysis pipeline for crosslinks and crosslink-spectrum-matches.

    Runs a standard down-stream analysis pipeline for crosslinks and crosslink-spectrum-matches. The pipeline first reads
    a result file and subsequently optionally filters the the read data for unique crosslinks and crosslink-spectrum-matches,
    optionally the data is validated by false discovery rate estimation and - also optionally - only target-target matches
    are returned. Internally the pipeline calls ``parser.read()``, ``transform.unique()``, ``transform.validate()``, and
    ``transform.targets_only()``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the result file(s) or a file-like object/stream.
    engine : "Custom", "MaxQuant", "MaxLynx", "MeroX", "MS Annika", "mzIdentML", "pLink", "Scout", "xiSearch/xiFDR", or "XlinkX"
        Crosslink search engine or format of the result file.
    crosslinker : str
        Name of the used cross-linking reagent, for example "DSSO".
    unique : dict of str, any, or bool, or None, default = True
        If ``transform.unique()`` should be run in the pipeline. If None or False this step is omitted.
        If True this step is run with default parameters. If a dictionary is given it should contain parameters for
        running ``transform.unique()``. Omitting a parameter in the dictionary will fall back to its default value.
    validate : dict of str, any, or bool, or None, default = True
        If ``transform.validate()`` should be run in the pipeline. If None or False this step is omitted.
        If True this step is run with default parameters. If a dictionary is given it should contain parameters for
        running ``transform.validate()``. Omitting a parameter in the dictionary will fall back to its default value.
    targets_only : bool, or None, default = True
        If ``transform.targets_only()`` should be run in the pipeline. If None or False this step is omitted.
    **kwargs
        Any additional parameters will be passed to the specific result file parsers.

    Returns
    -------
    dict of str, any
        The transformed parser_result after all pipeline steps are completed.

    Raises
    ------
    TypeError
        If any of the parameters do not have the correct type.

    Notes
    -----
    Various helpful pipeline information is also printed to ``stdout``.

    Examples
    --------
    >>> from pyXLMS.pipelines import pipeline
    >>> pr = pipeline(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ...     unique=True,
    ...     validate={"fdr": 0.05, "formula": "(TD-DD)/TT"},
    ...     targets_only=True,
    ... )
    Reading MS Annika CSMs...: 100%|██████████████████████████████████████████████████| 826/826 [00:00<00:00, 10337.98it/s]
    ---- Summary statistics before pipeline ----
    Number of CSMs: 826.0
    Number of unique CSMs: 826.0
    Number of intra CSMs: 803.0
    Number of inter CSMs: 23.0
    Number of target-target CSMs: 786.0
    Number of target-decoy CSMs: 39.0
    Number of decoy-decoy CSMs: 1.0
    Minimum CSM score: 1.11
    Maximum CSM score: 452.99
    Iterating over scores for FDR calculation...:   0%|                                            | 0/826 [00:00<?, ?it/s]
    ---- Summary statistics after pipeline ----
    Number of CSMs: 786.0
    Number of unique CSMs: 786.0
    Number of intra CSMs: 774.0
    Number of inter CSMs: 12.0
    Number of target-target CSMs: 786.0
    Number of target-decoy CSMs: 0.0
    Number of decoy-decoy CSMs: 0.0
    Minimum CSM score: 1.28
    Maximum CSM score: 452.99
    ---- Performed pipeline steps ----
    :: parser.read() ::
    :: parser.read() :: params :: <params omitted>
    :: transform.unique() ::
    :: transform.unique() :: params :: by=peptide
    :: transform.unique() :: params :: score=higher_better
    :: transform.validate() ::
    :: transform.validate() :: params :: fdr=0.05
    :: transform.validate() :: params :: formula=(TD-DD)/TT
    :: transform.validate() :: params :: score=higher_better
    :: transform.validate() :: params :: separate_intra_inter=False
    :: transform.validate() :: params :: ignore_missing_labels=False
    :: transform.targets_only() ::
    :: transform.targets_only() :: params :: no params
    """
    # steps: reading
    pr = read(files, engine=engine, crosslinker=crosslinker, **kwargs)
    # steps: summary (before)
    print("---- Summary statistics before pipeline ----")
    _ = transform_summary(pr)
    # steps: unique
    unique_params = {"by": "peptide", "score": "higher_better"}
    if unique is not None:
        if isinstance(unique, dict):
            unique_params.update(unique)
            pr = transform_unique(
                pr,
                by=str(unique_params["by"]),  # pyright: ignore[reportArgumentType]
                score=str(unique_params["score"]),  # pyright: ignore[reportArgumentType]
            )
        elif isinstance(unique, bool):
            if unique:
                pr = transform_unique(
                    pr,
                    by=str(unique_params["by"]),  # pyright: ignore[reportArgumentType]
                    score=str(unique_params["score"]),  # pyright: ignore[reportArgumentType]
                )
        else:
            raise TypeError(
                "Parameter unique has to be a dictionary of parameters for transform.unique(), a boolean or None!"
            )
    # steps: validate
    validate_params = {
        "fdr": 0.01,
        "formula": "D/T",
        "score": "higher_better",
        "separate_intra_inter": False,
        "ignore_missing_labels": False,
    }
    if validate is not None:
        if isinstance(validate, dict):
            validate_params.update(validate)
            pr = transform_validate(
                pr,
                fdr=float(validate_params["fdr"]),
                formula=str(validate_params["formula"]),  # pyright: ignore[reportArgumentType]
                score=str(validate_params["score"]),  # pyright: ignore[reportArgumentType]
                separate_intra_inter=bool(validate_params["separate_intra_inter"]),
                ignore_missing_labels=bool(validate_params["ignore_missing_labels"]),
            )
        elif isinstance(validate, bool):
            if validate:
                pr = transform_validate(
                    pr,
                    fdr=float(validate_params["fdr"]),
                    formula=str(validate_params["formula"]),  # pyright: ignore[reportArgumentType]
                    score=str(validate_params["score"]),  # pyright: ignore[reportArgumentType]
                    separate_intra_inter=bool(validate_params["separate_intra_inter"]),
                    ignore_missing_labels=bool(
                        validate_params["ignore_missing_labels"]
                    ),
                )
        else:
            raise TypeError(
                "Parameter validate has to be a dictionary of parameters for transform.validate(), a boolean or None!"
            )
    # steps: targets only
    if targets_only is not None:
        if isinstance(targets_only, bool):
            if targets_only:
                pr = transform_targets_only(pr)
        else:
            raise TypeError("Parameter targets_only has to be a boolean or None!")
    # steps: summary (after)
    print("---- Summary statistics after pipeline ----")
    _ = transform_summary(pr)
    # steps: pipeline summary
    print("---- Performed pipeline steps ----")
    print(":: parser.read() ::")
    print(":: parser.read() :: params :: <params omitted>")
    if unique is not None:
        if isinstance(unique, dict) or (isinstance(unique, bool) and unique):
            print(":: transform.unique() ::")
            for k, v in unique_params.items():
                print(f":: transform.unique() :: params :: {k}={v}")
    if validate is not None:
        if isinstance(validate, dict) or (isinstance(validate, bool) and validate):
            print(":: transform.validate() ::")
            for k, v in validate_params.items():
                print(f":: transform.validate() :: params :: {k}={v}")
    if targets_only is not None and targets_only:
        print(":: transform.targets_only() ::")
        print(":: transform.targets_only() :: params :: no params")
    # steps: finalize
    if not isinstance(pr, dict):
        raise RuntimeError(
            "Something went wrong while running the pipeline.\n"
            f"Expected data type: dict. Got: {type(pr)}."
        )
    return pr



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest
from typing import List, Dict, Any


def get_fdr_strict(data: List[Dict[str, Any]]) -> float:
    D = 0
    T = 0
    for item in data:
        if not item["alpha_decoy"] and not item["beta_decoy"]:
            T += 1
        else:
            D += 1
    return D / T


def get_fdr_relaxed(data: List[Dict[str, Any]]) -> float:
    D = 0
    DT = 0
    T = 0
    for item in data:
        if not item["alpha_decoy"] and not item["beta_decoy"]:
            T += 1
        elif item["alpha_decoy"] and item["beta_decoy"]:
            D += 1
        else:
            DT += 1
    if (DT - D) < 0.0:
        raise RuntimeError("Negative FDR!")
    return (DT - D) / T


def test1():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    csms = pr["crosslink-spectrum-matches"]
    csms = annotate_fdr(csms)
    validated_csms = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert len(validated_csms) == 705


def test2():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    pr = annotate_fdr(pr)
    validated_csms = [
        csm
        for csm in pr["crosslink-spectrum-matches"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert len(validated_csms) == 705
    validated_xls = [
        xl
        for xl in pr["crosslinks"]
        if xl["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert len(validated_xls) == 226


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 826
    csms = annotate_fdr(csms)
    assert len(csms) == 826
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert len(validated) == 705


def test4():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    pr = annotate_fdr(pr)
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for csm in pr["crosslink-spectrum-matches"]:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    for xl in pr["crosslinks"]:
        assert "pyXLMS_annotated_FDR" in xl["additional_information"]
    validated_csms = [
        csm
        for csm in pr["crosslink-spectrum-matches"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    validated_xls = [
        xl
        for xl in pr["crosslinks"]
        if xl["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert len(validated_csms) == 705
    assert len(validated_xls) == 226


def test5():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    pr = annotate_fdr(pr)
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for csm in pr["crosslink-spectrum-matches"]:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    for xl in pr["crosslinks"]:
        assert "pyXLMS_annotated_FDR" in xl["additional_information"]
    validated_csms = [
        csm
        for csm in pr["crosslink-spectrum-matches"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.05
    ]
    validated_xls = [
        xl
        for xl in pr["crosslinks"]
        if xl["additional_information"]["pyXLMS_annotated_FDR"] <= 0.05
    ]
    assert len(validated_csms) == 825
    assert len(validated_xls) == 260


def test6():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )

    err_str = (
        r"Parameter 'formula' has to be one of 'D\/T', '\(TD\+DD\)\/TT' or '\(TD\-DD\)\/TT'! Where D and DD is the number of decoys, T and TT the number of targets, "
        r"and TD the number of target-decoys!"
    )
    with pytest.raises(
        TypeError,
        match=err_str,
    ):
        _annotated = annotate_fdr(pr, formula="T/D")


def test7():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )

    err_str = r"Parameter 'score' has to be one of 'higher_better' or 'lower_better'!"
    with pytest.raises(
        TypeError,
        match=err_str,
    ):
        _annotated = annotate_fdr(pr, score="lower")


def test8():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/pyxlms/csm_min.txt",
        engine="custom",
        crosslinker="DSS",
    )

    err_str = (
        r"Can't annotate data if 'score' or target\/decoy labels are missing! Selecting 'ignore_missing_labels \= True' will ignore crosslinks and crosslink-spectrum-matches "
        r"that don't have a valid target\/decoy label and filter them out!"
    )
    with pytest.raises(
        ValueError,
        match=err_str,
    ):
        _annotated = annotate_fdr(pr)


def test9():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        "data/_test/validate/csms.txt",
        engine="custom",
        crosslinker="DSS",
    )

    err_str = r"Can't annotate FDR with formula '\(TD\-DD\)\/TT' when there are no TD matches! Please select the default formula instead!"
    with pytest.raises(
        ValueError,
        match=err_str,
    ):
        _annotated = annotate_fdr(pr, formula="(TD-DD)/TT")


def test10():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    pr = annotate_fdr(pr, formula="(TD+DD)/TT")
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for csm in pr["crosslink-spectrum-matches"]:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    for xl in pr["crosslinks"]:
        assert "pyXLMS_annotated_FDR" in xl["additional_information"]
    validated_csms = [
        csm
        for csm in pr["crosslink-spectrum-matches"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    validated_xls = [
        xl
        for xl in pr["crosslinks"]
        if xl["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_strict(validated_csms) < 0.01
    assert get_fdr_strict(validated_xls) < 0.01


def test11():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    pr["crosslinks"] = None
    pr = annotate_fdr(pr, formula="(TD-DD)/TT")
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert pr["crosslinks"] is None
    for csm in pr["crosslink-spectrum-matches"]:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated_csms = [
        csm
        for csm in pr["crosslink-spectrum-matches"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_relaxed(validated_csms) < 0.01


def test12():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for csm in pr["crosslink-spectrum-matches"]:
        csm["score"] = -csm["score"]
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD+DD)/TT",
        score="lower_better",
    )
    assert len(csms) == 826
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated_csms = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_strict(validated_csms) < 0.01


def test13():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for csm in pr["crosslink-spectrum-matches"]:
        csm["score"] = -csm["score"]
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD-DD)/TT",
        score="lower_better",
    )
    assert len(csms) == 826
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated_csms = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_relaxed(validated_csms) < 0.01


@pytest.mark.slow
def test14():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr
    from pyXLMS.transform import filter_crosslink_type

    pr = read(
        "data/_test/validate/csms_25000.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
        unsafe=True,
        verbose=0,
    )
    for csm in pr["crosslink-spectrum-matches"]:
        csm["score"] = -csm["score"]
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD+DD)/TT",
        score="lower_better",
        separate_intra_inter=True,
    )
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    separate = filter_crosslink_type(csms)
    intra = [
        csm
        for csm in separate["Intra"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    inter = [
        csm
        for csm in separate["Inter"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_strict(intra) < 0.01
    assert get_fdr_strict(inter) < 0.01


@pytest.mark.slow
def test15():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr
    from pyXLMS.transform import filter_crosslink_type

    pr = read(
        "data/_test/validate/csms_25000.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
        unsafe=True,
        verbose=0,
    )
    for csm in pr["crosslink-spectrum-matches"]:
        csm["score"] = -csm["score"]
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD-DD)/TT",
        score="lower_better",
        separate_intra_inter=True,
    )
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    separate = filter_crosslink_type(csms)
    intra = [
        csm
        for csm in separate["Intra"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    inter = [
        csm
        for csm in separate["Inter"]
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_relaxed(intra) < 0.01
    assert get_fdr_relaxed(inter) < 0.01


def test16():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for i, csm in enumerate(pr["crosslink-spectrum-matches"]):
        csm["score"] = -csm["score"]
        if i < 5:
            csm["alpha_decoy"] = None
        elif i < 10:
            csm["beta_decoy"] = None
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD+DD)/TT",
        score="lower_better",
        ignore_missing_labels=True,
    )
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated_csms = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_strict(validated_csms) < 0.01


def test17():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for i, csm in enumerate(pr["crosslink-spectrum-matches"]):
        csm["score"] = -csm["score"]
        if i < 5:
            csm["alpha_decoy"] = None
        elif i < 10:
            csm["beta_decoy"] = None
    csms = annotate_fdr(
        pr["crosslink-spectrum-matches"],
        formula="(TD-DD)/TT",
        score="lower_better",
        ignore_missing_labels=True,
    )
    for csm in csms:
        assert "pyXLMS_annotated_FDR" in csm["additional_information"]
    validated_csms = [
        csm
        for csm in csms
        if csm["additional_information"]["pyXLMS_annotated_FDR"] <= 0.01
    ]
    assert get_fdr_relaxed(validated_csms) < 0.01


def test18():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for i, csm in enumerate(pr["crosslink-spectrum-matches"]):
        csm["score"] = -csm["score"]
        if i < 5:
            csm["completeness"] = "partial"
            csm["alpha_decoy"] = None
        elif i < 10:
            csm["completeness"] = "partial"
            csm["beta_decoy"] = None
    err_str = (
        r"Can't annotate data if 'score' or target\/decoy labels are missing! Selecting 'ignore_missing_labels \= True' will ignore crosslinks and crosslink-spectrum-matches "
        r"that don't have a valid target\/decoy label and filter them out!"
    )
    with pytest.raises(
        ValueError,
        match=err_str,
    ):
        _annotated = annotate_fdr(
            pr["crosslink-spectrum-matches"],
            formula="(TD+DD)/TT",
            score="lower_better",
            ignore_missing_labels=False,
        )


def test19():
    from pyXLMS.parser import read
    from pyXLMS.transform import annotate_fdr

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 826
    assert len(pr["crosslinks"]) == 300
    for i, csm in enumerate(pr["crosslink-spectrum-matches"]):
        csm["score"] = -csm["score"]
        if i < 5:
            csm["completeness"] = "partial"
            csm["alpha_decoy"] = None
        elif i < 10:
            csm["completeness"] = "partial"
            csm["beta_decoy"] = None
    err_str = (
        r"Can't annotate data if 'score' or target\/decoy labels are missing! Selecting 'ignore_missing_labels \= True' will ignore crosslinks and crosslink-spectrum-matches "
        r"that don't have a valid target\/decoy label and filter them out!"
    )
    with pytest.raises(
        ValueError,
        match=err_str,
    ):
        _annotated = annotate_fdr(
            pr["crosslink-spectrum-matches"],
            formula="(TD-DD)/TT",
            score="lower_better",
            ignore_missing_labels=False,
        )
